"""Download useful datasets."""

# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

from . import audio_or_visual_speech
from . import camh_kf_fnirs_fingertapping
from . import block_speech_noise
from . import fnirs_motor_group
from . import snirf_with_aux
